#ifndef CDictionaryIntToInt_H
#define CDictionaryIntToInt_H

#include "onewaylist.h"
#include <istream>
#include <ostream>
#include <type_traits>
#include <string>
#include <stdexcept>
#include <iostream>

template<typename KeyType, typename ValueType>
class CDictionaryIntToInt
{
    OnewayList<KeyType> key;
    OnewayList<ValueType> val;
    int total = 0;
    
public:
    CDictionaryIntToInt() = default;
    CDictionaryIntToInt(KeyType key, ValueType val) {
        this->add(key, val);
    }
    CDictionaryIntToInt(const CDictionaryIntToInt&);
    
    bool add(KeyType, ValueType, char = 0);
    ValueType get(KeyType key) const {
        return this->val.get(find(key));
    }
    int find(KeyType, bool = 0) const;
    bool remove(KeyType);
    void removepos(int);
    
    CDictionaryIntToInt operator+(const CDictionaryIntToInt&) const;
    CDictionaryIntToInt operator-(const CDictionaryIntToInt&) const;
    CDictionaryIntToInt& operator=(const CDictionaryIntToInt&);
    
    template<typename K, typename V>
    friend std::istream& operator>>(std::istream&, CDictionaryIntToInt<K, V>&);
    
    template<typename K, typename V>
    friend std::ostream& operator<<(std::ostream&, const CDictionaryIntToInt<K, V>&);
    
    bool operator==(const CDictionaryIntToInt&) const;
    int gettotal() const {
        return total;
    }
    OnewayList<KeyType> getKeyList() const {
        return key;
    }    OnewayList<KeyType> getValList() const {
        return val;
    }
};
template<typename KeyType, typename ValueType>
int CDictionaryIntToInt<KeyType, ValueType>::find(KeyType key, bool soft) const {
    return this->key.find(key, soft);
}

template<typename KeyType, typename ValueType>
bool CDictionaryIntToInt<KeyType, ValueType>::add(KeyType key, ValueType val, char force) {
    if (!force && find(key, true) != -1) {
        if (force == 2) return false;
        throw std::runtime_error("Ключ вже зареєстровано");
    }
    this->key.append(key);
    this->val.append(val);
    total++;
    return true;
}

template<typename KeyType, typename ValueType>
bool CDictionaryIntToInt<KeyType, ValueType>::remove(KeyType key) {
    int pos = this->find(key);
    if (pos == -1) return false;
    total--;
    this->key.remove(pos);
    this->val.remove(pos);
    return true;
}

template<typename KeyType, typename ValueType>
void CDictionaryIntToInt<KeyType, ValueType>::removepos(int pos) {
    total--;
    this->key.remove(pos);
    this->val.remove(pos);
}

template<typename KeyType, typename ValueType>
CDictionaryIntToInt<KeyType, ValueType>::CDictionaryIntToInt(const CDictionaryIntToInt& other) {
    for (int i = 0; i < other.total; i++) {
        this->add(other.key.get(i), other.val.get(i), 1);
    }
}


template<typename KeyType, typename ValueType>
CDictionaryIntToInt<KeyType, ValueType> CDictionaryIntToInt<KeyType, ValueType>::operator+(const CDictionaryIntToInt& other) const {
    if (total == 0) return other;
    CDictionaryIntToInt returnme(other);
    for (int i = 0; i < total; i++) {
        KeyType key = this->key.get(i);
        if (other.find(key, 1) == -1) returnme.add(key, this->val.get(i));
    }
    return returnme;
}

template<typename KeyType, typename ValueType>
CDictionaryIntToInt<KeyType, ValueType> CDictionaryIntToInt<KeyType, ValueType>::operator-(const CDictionaryIntToInt& other) const {
    CDictionaryIntToInt returnme(*this);
    for (int i = 0; i < returnme.total; i++)
        if (other.find(returnme.key.get(i), 1) != -1) {
            returnme.removepos(i);
            i--;
        }
    return returnme;
}

template<typename KeyType, typename ValueType>
CDictionaryIntToInt<KeyType, ValueType>& CDictionaryIntToInt<KeyType, ValueType>::operator=(const CDictionaryIntToInt& other) {
    if (this == &other) return *this;
    key.~OnewayList();
    val.~OnewayList();
    total = 0;
    for (int i = 0; i < other.total; i++) {
        this->add(other.key.get(i), other.val.get(i), true);
    }
    return *this;
}

template<typename KeyType, typename ValueType>
bool CDictionaryIntToInt<KeyType, ValueType>::operator==(const CDictionaryIntToInt& other) const {
    if (this == &other) return true;
    if (total != other.total) return false;
    for (int i = 0; i < total; i++)
        if (other.get(this->key.get(i)) != this->val.get(i)) return false;
    return true;
}

template<typename KeyType, typename ValueType>
std::istream& operator>>(std::istream& in, CDictionaryIntToInt<KeyType, ValueType>& dict) {
    KeyType key;
    ValueType value;
    
    while (true) {
        if constexpr (std::is_same<KeyType, const char*>::value) {
            std::string keyStr;
            if (!(in >> keyStr >> value)) {
                break; // Break the loop if reading fails
            }
            key = keyStr.c_str();
        } else {
            if (!(in >> key >> value)) {
                break; // Break the loop if reading fails
            }
        }
        
        if (!dict.add(key, value)) {
            if constexpr (std::is_same<KeyType, const char*>::value) {
#ifdef REV
                throw std::runtime_error("Помилка додавання пари");
#else
                throw std::runtime_error("Ключ \"" + std::string(key) + "\" вже було введено");
#endif
            } else {
#ifdef REV
                throw std::runtime_error("Помилка додавання пари");
#else
                throw std::runtime_error("Ключ \"" + std::to_string(key) + "\" вже було введено");
#endif
            }
        }
    }
    
    return in;
}



template<typename KeyType, typename ValueType>
std::ostream& operator<<(std::ostream& out, const CDictionaryIntToInt<KeyType, ValueType>& dict) {
    for (int i = 0; i < dict.total; i++) out << dict.key.get(i) << " -> " << dict.val.get(i) << std::endl;
    return out;
}
#endif // CDictionaryIntToInt_H
